//  Copyright (C) 2018-2019  SEIKO EPSON CORPORATION
//
//  License: GPL-3.0+
//  Author : SEIKO EPSON CORPORATION
//
//  This file is part of the 'Epson Scan 2' package.
//  This package is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License or, at
//  your option, any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//  You ought to have received a copy of the GNU General Public License
//  along with this package.  If not, see <http://www.gnu.org/licenses/>.
//  
//  usbInterfaceImpl.cpp -- template and derived ESC/I ESC/I-2 protocol commands

#include "usbInterfaceImpl.h"
bool USBInterfaceImpl::sm_isInitialized = false;
int USBInterfaceImpl::sm_nDefaultTimeout = 60 * 1000;
int USBInterfaceImpl::sm_nConnectionCount = 0;
libusb_context* USBInterfaceImpl::sm_ctx = nullptr;
USBInterfaceImpl::USBInterfaceImpl()
: m_handle(nullptr)
, m_vid(0)
, m_pid(0)
, m_cfg(-1)
, m_if(-1)
, m_bKernelDriverDetached(false)
, m_bulk_in(-1)
, m_bulk_out(-1)
, m_delegate(nullptr)
{
    if (!sm_isInitialized){
        int res = libusb_init(&sm_ctx);
        sm_isInitialized = !res;
        if (res){
            sm_ctx = nullptr;
            throw std::runtime_error("unable to initialise USB support");
        }
        libusb_set_debug(sm_ctx, 3);
    }
    ++sm_nConnectionCount;
}
USBInterfaceImpl::~USBInterfaceImpl()
{
    Close();
    if (0 == --sm_nConnectionCount){
        if (sm_ctx)
        {
            libusb_exit(sm_ctx);
            sm_ctx = nullptr;
            sm_isInitialized = false;
        }
    }
}
void USBInterfaceImpl::SetDelegate(IInterfaceDelegate *delegate)
{
    m_delegate = delegate;
}
ESErrorCode USBInterfaceImpl::Init(UInt16 vid, UInt16 pid)
{
 libusb_device **haystack;
 ssize_t cnt = libusb_get_device_list(sm_ctx, &haystack);
 m_vid = m_pid = 0;
 for (ssize_t i = 0; i < cnt; i++){
  if ( IsMatchID(vid, pid, haystack[i]) ){
   m_vid = vid;
   m_pid = pid;
   break;
  }
 }
 libusb_free_device_list(haystack, 1);
 return ( (m_vid!=0 && m_pid!=0) ? kESErrorNoError : kESErrorFatalError);
}
bool USBInterfaceImpl::IsMatchID(UInt16 vid, UInt16 pid, libusb_device *dev)
{
 struct libusb_device_descriptor dd;
 int res = libusb_get_device_descriptor (dev, &dd);
 if (res || vid != dd.idVendor || pid != dd.idProduct){
  return false;
 }
 return true;
}
ESErrorCode USBInterfaceImpl::Open()
{
    const int RETRY_OPENDEVICE = 5;
    const unsigned int INTERVAL_RETRY = 250;
    if (m_vid ==0 || m_pid ==0){
        return kESErrorFatalError;
    }
    for (int nRetry = 0; nRetry < RETRY_OPENDEVICE; nRetry++) {
        m_handle = libusb_open_device_with_vid_pid(sm_ctx, m_vid, m_pid);
        if (m_handle) {
            break;
        }
        Sleep(INTERVAL_RETRY);
    }
    if (!m_handle){
        return kESErrorFatalError;
    }
    if (SetupInterface() != kESErrorNoError){
        return kESErrorDeviceOpenError;
    }
    return kESErrorNoError;
}
bool USBInterfaceImpl::IsOpened()
{
 return m_handle ? true : false;
}
void USBInterfaceImpl::Close(void)
{
 if(m_if != -1){
  libusb_release_interface(m_handle, m_if);
  if (m_bKernelDriverDetached){
   libusb_attach_kernel_driver(m_handle, m_if);
  }
  m_cfg = -1;
  m_if = -1;
  m_bulk_in = -1;
  m_bulk_out = -1;
 }
 if(m_handle){
  libusb_close(m_handle);
  m_handle = nullptr;
 }
}
ESErrorCode USBInterfaceImpl::SetupInterface()
{
    libusb_device* dev = libusb_get_device(m_handle);
    if (!dev){
        return kESErrorFatalError;
    }
    struct libusb_device_descriptor dd;
    int res = libusb_get_device_descriptor(dev, &dd);
    if (res){
        return kESErrorFatalError;
    }
    uint8_t nNumCfg = dd.bNumConfigurations;
    int nCuffentCfg;
    res = libusb_get_configuration(m_handle, &nCuffentCfg);
    if (res){
        return kESErrorFatalError;
    }
    for(uint8_t nCfgIndex = 0; nCfgIndex < nNumCfg; ++nCfgIndex){
        m_cfg = -1;
        struct libusb_config_descriptor *cd = nullptr;
        res = libusb_get_config_descriptor(dev, nCfgIndex, &cd);
        if (res){
            libusb_free_config_descriptor(cd);
            return kESErrorFatalError;
        }
        if (nCuffentCfg != cd->bConfigurationValue){
            res = libusb_set_configuration(m_handle, 1);
            if (res){
                libusb_free_config_descriptor(cd);
                return kESErrorFatalError;
            }
        }
        uint8_t nNumIf = cd->bNumInterfaces;
        for(uint8_t nIfIndex = 0; nIfIndex < cd->bNumInterfaces; ++nIfIndex){
            m_if = -1;
            for(uint8_t a = 0; a < cd->interface[nIfIndex].num_altsetting; ++a){
                const struct libusb_interface_descriptor *ifd = &cd->interface[nIfIndex].altsetting[a];
                if (ifd->bInterfaceClass != 0xFF ||
                    ifd->bInterfaceProtocol != 0xFF) {
                    continue;
                }
                bool bKernelDriverDetached = false;
                if ( libusb_kernel_driver_active(m_handle, ifd->bInterfaceNumber) == 1){
                    res = libusb_detach_kernel_driver(m_handle, ifd->bInterfaceNumber);
                    if (res) {
                        libusb_free_config_descriptor(cd);
                        return kESErrorFatalError;
                    }
                    bKernelDriverDetached = true;
                }
                res = libusb_claim_interface(m_handle, ifd->bInterfaceNumber);
                if (res){
                    if (bKernelDriverDetached){
                        libusb_attach_kernel_driver(m_handle, ifd->bInterfaceNumber);
                    }
                    libusb_free_config_descriptor(cd);
                    return kESErrorFatalError;
                }
                m_bulk_in = -1; m_bulk_out = -1;
                for (int n = 0; n < ifd->bNumEndpoints; ++n) {
                    const struct libusb_endpoint_descriptor *ep = &ifd->endpoint[n];
                    if (LIBUSB_TRANSFER_TYPE_BULK == (LIBUSB_TRANSFER_TYPE_MASK & ep->bmAttributes)){
                        if (LIBUSB_ENDPOINT_DIR_MASK & ep->bEndpointAddress) {
                            m_bulk_in = ep->bEndpointAddress;
                        } else{
                            m_bulk_out = ep->bEndpointAddress;
                        }
                    }
                }
                if (m_bulk_in != -1 && m_bulk_out != -1){
                    m_if = ifd->bInterfaceNumber;
                    m_bKernelDriverDetached = bKernelDriverDetached;
                    break;
                }
                libusb_release_interface(m_handle, ifd->bInterfaceNumber);
                if (bKernelDriverDetached){
                    libusb_attach_kernel_driver(m_handle, ifd->bInterfaceNumber);
                }
            }
            if (m_if != -1){
                m_cfg = cd->bConfigurationValue;
                break;
            }
        }
        libusb_free_config_descriptor(cd);
        if (m_cfg != -1){
            break;
        }
    }
    return ( (m_if!=-1 && m_bulk_in!=-1 && m_bulk_out!=-1) ? kESErrorNoError : kESErrorFatalError);
}
ESErrorCode USBInterfaceImpl::Read(PESByte outBytes, UInt32 length)
{
    ESErrorCode ret = kESErrorNoError;
    int transferred;
    int err = libusb_bulk_transfer (m_handle, m_bulk_in, outBytes, length,
                                    &transferred, sm_nDefaultTimeout);
    if (LIBUSB_ERROR_PIPE == err) {
        err = libusb_clear_halt (m_handle, m_bulk_in);
    }
    if (err){
        ret = kESErrorDataReceiveFailure;
    }
    return ret;
}
ESErrorCode USBInterfaceImpl::Write(PESByte inBytes, UInt32 length)
{
    ESErrorCode ret = kESErrorNoError;
    int transferred;
    int err = libusb_bulk_transfer (m_handle, m_bulk_out, inBytes, length,
                                    &transferred, sm_nDefaultTimeout);
    if (LIBUSB_ERROR_PIPE == err){
        err = libusb_clear_halt (m_handle, m_bulk_out);
    }
    if (err){
        ret = kESErrorDataSendFailure;
    }
    return ret;
}
ESErrorCode USBInterfaceImpl::Write(PESByte inBytes, UInt32 length, UInt32 committedReadLength)
{
    return Write(inBytes, length);
}
bool USBInterfaceImpl::IsInterruptSupported()
{
 return false;
}
bool USBInterfaceImpl::IsSupportsExtendedTransfer()
{
    return false;
}
ESErrorCode USBInterfaceImpl::StartExtendedTransferWithBlocks(UInt32 blocks, UInt32 bytesPerBlock, UInt32 bytesOfLastBlock)
{
 return kESErrorFatalError;
}
ESErrorCode USBInterfaceImpl::ReadExtendedTransfer(PESByte theBytes, UInt32 length)
{
 return kESErrorFatalError;
}
ESErrorCode USBInterfaceImpl::AbortExtendedTransferWithCode(UInt8 code)
{
 return kESErrorFatalError;
}
bool USBInterfaceImpl::IsAvoidsAckWhileImageTransfer()
{
 return false;
}
int USBInterfaceImpl::WaitTimeAfterFinalize()
{
 return 0;
}
